from __future__ import division
"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["SslConfig", "BrokerConnection"]
from functools import partial
import logging
import ssl
import struct
import time

from .exceptions import SocketDisconnectedError
from .utils.socket import recvall_into
from .utils.compat import buffer

log = logging.getLogger(__name__)


class SslConfig(object):
    """Config object for SSL connections

    This aims to pick optimal defaults for the majority of use cases.  If you
    have special requirements (eg. you want to enable hostname checking), you
    may monkey-patch `self._wrap_socket` (see `_legacy_wrap_socket()` for an
    example) before passing the `SslConfig` to `KafkaClient` init, like so:

        config = SslConfig(cafile='/your/ca/file')
        config._wrap_socket = config._legacy_wrap_socket()
        client = KafkaClient('localhost:<ssl-port>', ssl_config=config)

    Alternatively, completely supplanting this class with your own is also
    simple: if you are not going to be using the `pykafka.rdkafka` classes,
    only a method `wrap_socket()` is expected (so you can eg. simply pass in
    a plain `ssl.SSLContext` instance instead).  The `pykafka.rdkafka`
    classes require four further attributes: `cafile`, `certfile`, `keyfile`,
    and `password` (the `SslConfig.__init__` docstring explains their meaning)
    """
    def __init__(self,
                 cafile,
                 certfile=None,
                 keyfile=None,
                 password=None):
        """Specify certificates for SSL connection

        :param cafile: Path to trusted CA certificate
        :type cafile: str
        :param certfile: Path to client certificate
        :type certfile: str
        :param keyfile: Path to client private-key file
        :type keyfile: str
        :param password: Password for private key
        :type password: bytes
        """
        self.cafile = cafile
        self.certfile = certfile
        self.keyfile = keyfile
        self.password = password
        self._wrap_socket = None

    def wrap_socket(self, sock):
        """Wrap a socket in an SSL context (see `ssl.wrap_socket`)

        :param socket: Plain socket
        :type socket: :class:`socket.socket`
        """
        if self._wrap_socket is None:
            if hasattr(ssl, 'SSLContext'):
                ssl_context = ssl.create_default_context(cafile=self.cafile)
                ssl_context.check_hostname = False
                if self.certfile is not None:
                    ssl_context.load_cert_chain(certfile=self.certfile,
                                                keyfile=self.keyfile,
                                                password=self.password)
                self._wrap_socket = ssl_context.wrap_socket
            else:
                self._wrap_socket = self._legacy_wrap_socket()
        return self._wrap_socket(sock)

    def _legacy_wrap_socket(self):
        """Create socket-wrapper on a pre-2.7.9 Python interpreter"""
        log.warning("SSL: using legacy fallback, which may use sub-optimal "
                    "defaults. Check that this satisfies your needs.")
        return partial(ssl.wrap_socket,
                       keyfile=self.keyfile,
                       certfile=self.certfile,
                       cert_reqs=ssl.CERT_REQUIRED,
                       # protocol-wise, best we can do pre-2.7.9:
                       ssl_version=ssl.PROTOCOL_TLSv1,
                       ca_certs=self.cafile)


class BrokerConnection(object):
    """
    BrokerConnection thinly wraps a `socket.create_connection` call
    and handles the sending and receiving of data that conform to the
    kafka binary protocol over that socket.
    """
    def __init__(self,
                 host,
                 port,
                 handler,
                 buffer_size=1024 * 1024,
                 source_host='',
                 source_port=0,
                 ssl_config=None):
        """Initialize a socket connection to Kafka.

        :param host: The host to which to connect
        :type host: str
        :param port: The port on the host to which to connect.  Assumed to be
            an ssl-endpoint if (and only if) `ssl_config` is also provided
        :type port: int
        :param handler: The :class:`pykafka.handlers.Handler` instance to use when
            creating a connection
        :type handler: :class:`pykafka.handlers.Handler`
        :param buffer_size: The size (in bytes) of the buffer in which to
            hold response data.
        :type buffer_size: int
        :param source_host: The host portion of the source address for
            the socket connection
        :type source_host: str
        :param source_port: The port portion of the source address for
            the socket connection
        :type source_port: int
        :param ssl_config: Config object for SSL connection
        :type ssl_config: :class:`pykafka.connection.SslConfig`
        """
        self._buff = bytearray(buffer_size)
        self.host = host
        self.port = port
        self._handler = handler
        self._socket = None
        self.source_host = source_host
        self.source_port = source_port
        self._wrap_socket = (
            ssl_config.wrap_socket if ssl_config else lambda x: x)

    def __del__(self):
        """Close this connection when the object is deleted."""
        self.disconnect()

    @property
    def connected(self):
        """Returns true if the socket connection is open."""
        return self._socket is not None

    def connect(self, timeout, attempts=1):
        """Connect to the broker, retrying if specified."""
        log.debug("Connecting to %s:%s", self.host, self.port)
        for attempt in range(0, attempts):
            try:
                self._socket = self._wrap_socket(
                    self._handler.Socket.create_connection(
                        (self.host, self.port),
                        timeout / 1000,
                        (self.source_host, self.source_port)
                    )
                )
                log.debug("Successfully connected to %s:%s", self.host, self.port)
                return
            except (self._handler.SockErr, self._handler.GaiError) as err:
                log.info("Attempt %s: failed to connect to %s:%s", attempt, self.host, self.port)
                log.info(err)
                if attempts > 1:
                    log.info("Retrying in 300ms.")
                    time.sleep(.3)
                continue

        raise SocketDisconnectedError("<broker {}:{}>".format(self.host, self.port))

    def disconnect(self):
        """Disconnect from the broker."""
        if self._socket is None:
            return
        try:
            self._socket.close()
        except IOError:
            pass
        finally:
            self._socket = None

    def reconnect(self):
        """Disconnect from the broker, then reconnect"""
        self.disconnect()
        self.connect(10 * 1000)

    def request(self, request):
        """Send a request over the socket connection"""
        bytes_ = request.get_bytes()
        if not self._socket:
            raise SocketDisconnectedError("<broker {}:{}>".format(self.host, self.port))
        try:
            self._socket.sendall(bytes_)
        except self._handler.SockErr as e:
            log.error("Failed to send data, error: %s" % repr(e))
            self.disconnect()
            raise SocketDisconnectedError("<broker {}:{}>".format(self.host, self.port))

    def response(self):
        """Wait for a response from the broker"""
        size = bytes()
        expected_len = 4  # Size => int32
        while len(size) != expected_len:
            try:
                r = self._socket.recv(expected_len - len(size))
            except IOError:
                r = None
            if r is None or len(r) == 0:
                # Happens when broker has shut down
                self.disconnect()
                raise SocketDisconnectedError("<broker {}:{}>".format(self.host, self.port))
            size += r
        size = struct.unpack('!i', size)[0]
        try:
            recvall_into(self._socket, self._buff, size)
        except SocketDisconnectedError:
            self.disconnect()
            raise SocketDisconnectedError("<broker {}:{}>".format(self.host, self.port))
        # Drop CorrelationId => int32
        return buffer(self._buff[4:4 + size])



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["RandomPartitioner", "BasePartitioner", "HashingPartitioner",
           "hashing_partitioner", "GroupHashingPartitioner"]
import random

from hashlib import sha1


class BasePartitioner(object):
    """Base class for custom class-based partitioners.

    A partitioner is used by the :class:`pykafka.producer.Producer` to
    decide which partition to which to produce messages.
    """
    def __call__(self, partitions, key=None):
        raise NotImplementedError('Subclasses must define their own '
                                  ' partitioner implementation')


class RandomPartitioner(BasePartitioner):
    """Returns a random partition out of all of the available partitions.

    Uses a non-random incrementing counter to provide even distribution across partitions
    without wasting CPU cycles
    """
    def __init__(self):
        self.idx = 0

    def __call__(self, partitions, key):
        self.idx = (self.idx + 1) % len(partitions)
        return partitions[self.idx]


class HashingPartitioner(BasePartitioner):
    """
    Returns a (relatively) consistent partition out of all available partitions
    based on the key.

    Messages that are published with the same keys are not guaranteed to end up
    on the same broker if the number of brokers changes (due to the addition
    or removal of a broker, planned or unplanned) or if the number of topics
    per partition changes. This is also unreliable when not all brokers are
    aware of a topic, since the number of available partitions will be in flux
    until all brokers have accepted a write to that topic and have declared how
    many partitions that they are actually serving.
    """
    def __init__(self, hash_func=None):
        """
        :param hash_func: hash function (defaults to :func:`hash`), should return
            an `int`. If hash randomization (Python 2.7) is enabled, a custom
            hashing function should be defined that is consistent between
            interpreter restarts.
        :type hash_func: function
        """
        self.hash_func = hash_func
        if self.hash_func is None:
            self.hash_func = lambda k: int(sha1(k).hexdigest(), 16)

    def __call__(self, partitions, key):
        """
        :param partitions: The partitions from which to choose
        :type partitions: sequence of :class:`pykafka.base.BasePartition`
        :param key: Key used for routing
        :type key: Any hashable type if using the default :func:`hash`
            implementation, any valid value for your custom hash function
        :returns: A partition
        :rtype: :class:`pykafka.base.BasePartition`
        """
        if key is None:
            raise ValueError(
                'key cannot be `None` when using hashing partitioner'
            )
        partitions = sorted(partitions)  # sorting is VERY important
        return partitions[abs(self.hash_func(key)) % len(partitions)]


hashing_partitioner = HashingPartitioner()


class GroupHashingPartitioner(BasePartitioner):
    """
    Messages published with the identical keys will be directed to a consistent subset of 'n'
    partitions from the set of available partitions. For example, if there are 16 partitions and group_size=4,
    messages with the identical keys will be shared equally between a subset of four partitions, instead of always
    being directed to the same partition.

    The same guarantee caveats apply as to the :class:`pykafka.base.HashingPartitioner`.
    """
    def __init__(self, hash_func, group_size=1):
        """
        :param hash_func: A hash function
        :type hash_func: function
        :param group_size: Size of the partition group to assign to. For example, if there are 16 partitions, and we
            want to smooth the distribution of identical keys between a set of 4, use 4 as the group_size.
        :type group_size: Integer value between (0, total_partition_count)
        """
        self.hash_func = hash_func
        self.group_size = group_size
        if self.hash_func is None:
            raise ValueError(
                'hash_func must be specified when using GroupHashingPartitioner'
            )
        if self.group_size < 1:
            raise ValueError(
                'group_size cannot be < 1 when using GroupHashingPartitioner'
            )

    def __call__(self, partitions, key):
        """
        :param partitions: The partitions from which to choose
        :type partitions: sequence of :class:`pykafka.base.BasePartition`
        :param key: Key used for routing
        :type key: Any hashable type if using the default :func:`hash`
            implementation, any valid value for your custom hash function
        :returns: A partition
        :rtype: :class:`pykafka.base.BasePartition`
        """
        if key is None:
            raise ValueError(
                'key cannot be `None` when using hashing partitioner'
            )
        if self.group_size > len(partitions):
            raise ValueError(
                'group_size cannot be > available partitions'
            )
        partitions = sorted(partitions)
        return partitions[abs(self.hash_func(key) + random.randrange(0, self.group_size)) % len(partitions)]



import unittest2

from pykafka.utils import struct_helpers


class StructHelpersTests(unittest2.TestCase):
    def test_basic_unpack(self):
        output = struct_helpers.unpack_from(
            'iiqhi',
            b'\x00\x00\x00\x01\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\n\x00<\x00\x00\x00\x04'
        )
        self.assertEqual(output, (1, 10, 10, 60, 4))

    def test_string_encoding(self):
        output = struct_helpers.unpack_from('S', b'\x00\x04test')
        self.assertEqual(output, (b'test',))

    def test_bytearray_unpacking(self):
        output = struct_helpers.unpack_from('Y', b'\x00\x00\x00\x04test')
        self.assertEqual(output, (b'test',))

    def test_array_unpacking(self):
        output = struct_helpers.unpack_from(
            '[i]',
            b'\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04'
        )
        # A 1-length tuple with a 4-length tuple as the element
        self.assertEqual(output, [1, 2, 3, 4])

if __name__ == '__main__':
    unittest2.main()



"""
Author: Keith Bourgoin
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["encode_gzip", "decode_gzip", "encode_snappy", "decode_snappy", "encode_lz4", "decode_lz4", "encode_lz4_old_kafka","decode_lz4_old_kafka"]
import gzip
from io import BytesIO
import logging
import struct

from .compat import range, buffer, IS_PYPY, PY3

try:
    import snappy
except ImportError:
    snappy = None

try:
    import lz4.frame as lz4
except ImportError:
    lz4 = None

try:
    import lz4f
except ImportError:
    lz4f = None

try:
    import xxhash
except ImportError:
    xxhash = None

log = logging.getLogger(__name__)
# constants used in snappy xerial encoding/decoding
_XERIAL_V1_HEADER = (-126, b'S', b'N', b'A', b'P', b'P', b'Y', 0, 1, 1)
_XERIAL_V1_FORMAT = 'bccccccBii'


def encode_gzip(buff):
    """Encode a buffer using gzip"""
    sio = BytesIO()
    f = gzip.GzipFile(fileobj=sio, mode="w")
    f.write(buff)
    f.close()
    sio.seek(0)
    output = sio.read()
    sio.close()
    return output


def decode_gzip(buff):
    """Decode a buffer using gzip"""
    sio = BytesIO(buff)
    f = gzip.GzipFile(fileobj=sio, mode='r')
    output = f.read()
    f.close()
    sio.close()
    return output


def encode_snappy(buff, xerial_compatible=False, xerial_blocksize=32 * 1024):
    """Encode a buffer using snappy

    If xerial_compatible is set, the buffer is encoded in a fashion compatible
    with the xerial snappy library.

    The block size (xerial_blocksize) controls how frequently the blocking
    occurs. 32k is the default in the xerial library.

    The format is as follows:
    +-------------+------------+--------------+------------+--------------+
    |   Header    | Block1 len | Block1 data  | Blockn len | Blockn data  |
    |-------------+------------+--------------+------------+--------------|
    |  16 bytes   |  BE int32  | snappy bytes |  BE int32  | snappy bytes |
    +-------------+------------+--------------+------------+--------------+

    It is important to note that `blocksize` is the amount of uncompressed
    data presented to snappy at each block, whereas `blocklen` is the
    number of bytes that will be present in the stream.

    Adapted from kafka-python
    https://github.com/mumrah/kafka-python/pull/127/files
    """
    #snappy segfaults if it gets a read-only buffer on PyPy
    if IS_PYPY or PY3:
        buff = bytes(buff)
    if snappy is None:
        raise ImportError("Please install python-snappy")
    if xerial_compatible:
        def _chunker():
            for i in range(0, len(buff), xerial_blocksize):
                yield buff[i:i + xerial_blocksize]
        out = BytesIO()
        full_data = list(zip(_XERIAL_V1_FORMAT, _XERIAL_V1_HEADER))
        header = b''.join(
            [struct.pack('!' + fmt, dat) for fmt, dat in full_data]
        )

        out.write(header)
        for chunk in _chunker():
            block = snappy.compress(chunk)
            block_size = len(block)
            out.write(struct.pack('!i', block_size))
            out.write(block)
        out.seek(0)
        return out.read()
    else:
        return snappy.compress(buff)


def decode_snappy(buff):
    """Decode a buffer using Snappy

    If xerial is found to be in use, the buffer is decoded in a fashion
    compatible with the xerial snappy library.

    Adapted from kafka-python
    https://github.com/mumrah/kafka-python/pull/127/files
    """
    if snappy is None:
        raise ImportError("Please install python-snappy")
    if _detect_xerial_stream(buff):
        out = BytesIO()
        body = buffer(buff[16:])
        if PY3:  # workaround for snappy bug
            body = bytes(body)
        length = len(body)
        cursor = 0
        while cursor < length:
            block_size = struct.unpack_from('!i', body[cursor:])[0]
            cursor += 4
            end = cursor + block_size
            out.write(snappy.decompress(body[cursor:end]))
            cursor = end
        out.seek(0)
        return out.read()
    else:
        return snappy.decompress(buff)


def _detect_xerial_stream(buff):
    """Detects the use of the xerial snappy library

    Returns True if the data given might have been encoded with the blocking
    mode of the xerial snappy library.

    This mode writes a magic header of the format:
        +--------+--------------+------------+---------+--------+
        | Marker | Magic String | Null / Pad | Version | Compat |
        |--------+--------------+------------+---------+--------|
        |  byte  |   c-string   |    byte    |  int32  | int32  |
        |--------+--------------+------------+---------+--------|
        |  -126  |   'SNAPPY'   |     \0     |         |        |
        +--------+--------------+------------+---------+--------+

    `pad` appears to be to ensure that SNAPPY is a valid c-string.
    `version` is the version of this format as written by xerial.
    In the wild, this is currently 1, and as such we only support v1.

    `compat` is there to claim the miniumum supported version that
    can read a xerial block stream; presently in the wild this is 1.

    Adapted from kafka-python
    https://github.com/mumrah/kafka-python/pull/127/files
    """
    if len(buff) > 16:
        header = struct.unpack('!' + _XERIAL_V1_FORMAT, bytes(buff)[:16])
        return header == _XERIAL_V1_HEADER
    return False


if lz4:
    encode_lz4 = lz4.compress  # pylint: disable-msg=no-member
elif lz4f:
    encode_lz4 = lz4f.compressFrame  # pylint: disable-msg=no-member
else:
    encode_lz4 = None


def decode_lz4f(buff):
    """Decode payload using interoperable LZ4 framing. Requires Kafka >= 0.10"""
    # pylint: disable-msg=no-member
    ctx = lz4f.createDecompContext()
    data = lz4f.decompressFrame(buff, ctx)
    lz4f.freeDecompContext(ctx)

    # lz4f python module does not expose how much of the payload was
    # actually read if the decompression was only partial.
    if data['next'] != 0:
        raise RuntimeError('lz4f unable to decompress full buffer')
    return data['decomp']


if lz4:
    decode_lz4 = lz4.decompress  # pylint: disable-msg=no-member
elif lz4f:
    decode_lz4 = decode_lz4f
else:
    decode_lz4 = None


def encode_lz4_old_kafka(buff):
    """Encode buff for 0.8/0.9 brokers -- requires an incorrect header checksum.

    Reference impl: https://github.com/dpkp/kafka-python/blob/a00f9ead161e8b05ac953b460950e42fa0e0b7d6/kafka/codec.py#L227
    """
    assert xxhash is not None
    data = encode_lz4(buff)
    header_size = 7
    flg = data[4]
    if not isinstance(flg, int):
        flg = ord(flg)

    content_size_bit = ((flg >> 3) & 1)
    if content_size_bit:
        # Old kafka does not accept the content-size field
        # so we need to discard it and reset the header flag
        flg -= 8
        data = bytearray(data)
        data[4] = flg
        data = bytes(data)
        buff = data[header_size+8:]
    else:
        buff = data[header_size:]

    # This is the incorrect hc
    hc = xxhash.xxh32(data[0:header_size-1]).digest()[-2:-1]  # pylint: disable-msg=no-member

    return b''.join([
        data[0:header_size-1],
        hc,
        buff
    ])


def decode_lz4_old_kafka(buff):
    """Decode buff for 0.8/0.9 brokers

    Reference impl: https://github.com/dpkp/kafka-python/blob/a00f9ead161e8b05ac953b460950e42fa0e0b7d6/kafka/codec.py#L258
    """
    assert xxhash is not None
    # Kafka's LZ4 code has a bug in its header checksum implementation
    header_size = 7
    if isinstance(buff[4], int):
        flg = buff[4]
    else:
        flg = ord(buff[4])
    content_size_bit = ((flg >> 3) & 1)
    if content_size_bit:
        header_size += 8

    # This should be the correct hc
    hc = xxhash.xxh32(buff[4:header_size-1]).digest()[-2:-1]  # pylint: disable-msg=no-member

    munged_buff = b''.join([
        buff[0:header_size-1],
        hc,
        buff[header_size:]
    ])
    return decode_lz4(munged_buff)



import unittest
from uuid import uuid4

from pykafka import KafkaClient, Topic
from pykafka.utils.compat import itervalues
from pykafka.test.utils import get_cluster, stop_cluster


class ClusterIntegrationTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kafka = get_cluster()
        cls.client = KafkaClient(cls.kafka.brokers)

    @classmethod
    def tearDownClass(cls):
        stop_cluster(cls.kafka)

    def test_topic_autocreate(self):
        topic_name = uuid4().hex.encode()
        topic = self.client.topics[topic_name]
        self.assertTrue(isinstance(topic, Topic))

    def test_exclude_internal_topics(self):
        """Test exclude_internal_topics setting

        See also #277 for a related bug.
        """
        topic_name = b"__starts_with_underscores"
        with self.assertRaises(KeyError):
            topic = self.client.topics[topic_name]

        client = KafkaClient(self.kafka.brokers, exclude_internal_topics=False)
        topic = client.topics[topic_name]
        self.assertTrue(isinstance(topic, Topic))

    def test_topic_updates(self):
        startlen = len(self.client.topics)
        name_a = uuid4().hex.encode()
        name_b = uuid4().hex.encode()

        self.client.topics[name_a]
        self.client.topics[name_a]  # 2nd time shouldn't affect len(topics)
        self.client.topics[name_b]
        self.assertEqual(len(self.client.topics), startlen + 2)
        self.assertIn(name_a, self.client.topics)

        self.kafka.delete_topic(name_a)
        self.client.update_cluster()
        self.assertEqual(len(self.client.topics), startlen + 1)
        self.assertNotIn(name_a, self.client.topics)
        self.assertIn(name_b, self.client.topics)

        self.kafka.delete_topic(name_b)
        self.client.update_cluster()
        self.assertEqual(len(self.client.topics), startlen)
        self.assertNotIn(name_b, self.client.topics)

    def test_zk_connect(self):
        """Clusters started with broker lists and zk connect strings should get same brokers"""
        zk_client = KafkaClient(zookeeper_hosts=self.kafka.zookeeper)
        kafka_client = KafkaClient(hosts=self.kafka.brokers)
        zk_brokers = ["{}:{}".format(b.host, b.port)
                      for b in itervalues(zk_client.brokers)]
        kafka_brokers = ["{}:{}".format(b.host, b.port)
                         for b in itervalues(kafka_client.brokers)]
        self.assertEqual(zk_brokers, kafka_brokers)

if __name__ == "__main__":
    unittest.main()
